c
c  Arguement list has been expanded to include ro
c  Feb 22, 1999,  BJS

      subroutine dgmr(n,im,neig,ip,maxits,iout,ioutEig,eps,
     ;	rhs,sol,i,its,vv,nvv,icode,ipo,ipi,ro)
c
      implicit none  
      integer n,im,neig,ip,maxits,iout,ioutEig,i,its,nvv,icode,ipo,ipi
      real*8  eps,rhs(n),sol(n),vv(n,*)
c
c-----------------------------------------------------------------------
c -----DGMR-----
c Deflated GMRES with right preconditioning. The subroutine deigen 
c is called to estimate the eigenvectors corresponding to the 
c eigenvalues of smallest magnitude. dgmr uses reverse communication 
c for matrix vector products and preconditioning.

c References: 
c -R B Morgan 'A restarted GMRES method augmented with eigenvectors'
c  SIAM J. Matrix Analysis and Applications , 16(4):1154-1171,1995.
c -Andrew Chapman and Yousef Saad 'Deflated and augmented Krylov
c  subspace techniques' MSI report UMSI 95/181 September 1995
c
c In the references above, eigenvectors are added as the last vectors of
c the augmented Krylov subspace. Here they are added as the first 
c vectors. This usually results in better convergence between restarts.
c
c-----------------------------------------------------------------------
c ---on entry---
c n      = dimension of matrix 
c im     = max number vectors in Krylov subspace (including eigenvectors)
c neig   = number of eigenvectors to be calculated and used in deflation.
c          Set neig=0 for no deflation, and 1<=neig<=(im-1) for deflation. 
c 	   Usually neig<<(im-1), and a typical value is neig=4. See notes 
c  	   on ip and neig below.
c ip     = number of eigenvectors available on calling bgmr. Set ip=0
c          unless dgmr is used to solve for for multiple rhs. Here 
c 	   set ip=0 for first rhs, and dgmr will automatically set 
c 	   ip=neig for all other rhs. This causes eigenvectors to be 
c 	   recycled, and used as starting eigenvectors when solving 
c 	   for subsequent rhs. See notes on ip and neig below
c maxits = max number of total gmres  steps allowed
c iout   = number of file for writing output, iout=0 ---> no output.
c	   This file needs to be opened in the calling program.
c ioutEig= number of file for writing eigenvalue/vector residual. if
c          ioutEig = 0, no eigenvalue/vector residuals are calculated
c          or written. Note that calculating the residuals is time
c          consuming, and they are they are calculated only to be
c          output in ioutEig, they are never used. Set ioutEig=0, unless
c          you are interested in seeing the residual in the calculated
c          eigenvectors. This file needs to be opened in the calling
c          program
c eps    = tolerance parameter, dgmr stops if ro_{its}.le.ro_{0}*eps
c	   where ro is the residual norm |rhs-A sol|, and its is the 
c	   step number 
c rhs    = right-hand-side vector, dimension n
c sol    = initial guess of solution on input, dimension n
c vv     = array used to store Krylov subspace V_{m+1} at vv(1,1),
c          eigenvectors at vv(1,ifirst), and workspace at vv(1,iwork). In 
c  	   calling program, vv must be an array of size >= 
c  	   n * ( im+1 + max(neig,ip) + 2 ) . For more information, see 
c 	   notes on V_{m+1} and W_{m} below
c nvv    = number entries in vv in calling program, used to check size
c
c ---on return--- 
c sol    = solution on return -- 
c i      = step number in current Arnoldi .. Should not be changed 
c its	 = total number of steps, including restarts
c icode  = used for reverse communication protocol. See notes below
c ipo,ipi= pointers for reverse communication. See notes below
c ip     = If solving for multiple rhs, ip is the number of eigenvectors 
c 	   available as starting eigenvectors when solving for the next 
c 	   rhs
c vv     = contains ip eigenvectors at vv(1,ifirst), vv(1,ifirst+1),
c          ... vv(1,ifirst-1+ip)
c
c-----------------------------------------------------------------------
c REVERSE COMMUNICATION 
c----------------------
c USAGE: DGMR should be put in a loop and the loop should be 
c active for as long as icode is not equal to 0. On return dgmr 
c will ask the user to perform an operation on vv(ipo), and put the 
c result in vv(ipi) The operation could be a matrix vector 
c multiplication, or a preconditioning operation. After the operation, 
c call DGMR again. This should be done for as long as icode .gt. 0. 
c On entry icode should always be equal to 0. 
c * if icode  =  1 --> operation requested is preconditioning 
c * if icode  =  2 --> operation requested is a matrix-vector product. 
c * if icode  =  0 --> termination criterion satisfied: residual reduced
c		       below eps 
c * if icode  = -1 --> unsuccessful termination: residual not reduced
c		       below eps in maxits steps
c
c Here is a typical way of running dgmr: 
c
c     icode = 0
c100  continue
c     call dgmr(n,im,neig,ip,maxits,iout,ioutEig,eps,
c    ;  rhs,sol,i,its,vv,nvv,icode,ipo,ipi)
cc
c     if(icode.eq.0)then
c         write(*,"( ' dgmr converged in',I5,' iterations')")its
c     elseif (icode .eq. -1)then
c         write(*,"( ' ***no convergence in',I5,' gmres its')")its
c     elseif (icode .eq. 2) then
c         ---matrix vector multiply
c         call amux(ndim,vv(ipo),vv(ipi),a,ja,ia)
c         goto 100
c     else if (icode .eq. 1) then
c         ---preconditioning
c         call precon(ndim,vv(ipo),vv(ipi), ... )
c         go to 100
c     else
c         stop '*** in dgmr, icode must equal -1,0,1,2 ***'
c     endif
c         ----- done ---- 
c         .........
c Here the subroutine amux performs the matrix vector product
c	vv(ipi) = A vv(ipo)
c the subroutine precon performs the right preconditioning operation
c	vv(ipi) = M^{-1} vv(ipo)
c-----------------------------------------------------------------------
c Notes on ip and neig for single rhs
c -----------------------------------
c When solving for a single rhs, dgmr calculates neig eigenvectors
c at each restart, and adds them as the first ip vectors of the next
c Krylov subspace. When dgmr is first called, no eigenvectors are 
c available, and ip=0. After the first restart, eigenvectors are 
c available, and  dgmr automatically sets ip=neig. 
c
c If this code is used for a single rhs only, change the following
c statement just before the call to deigen
c     if( neig.gt.0 .and. ((.not.conv).or.(icol.gt.neig)) )then
c to the statement
c     if( neig.gt.0 .and. (.not.conv) )then
c This tells the code that it is not necessary to calculate eigenvectors
c at convergence, because they are not going to be used to solve for
c another rhs
c
c Notes on ip and neig for multiple rhs
c -------------------------------------
c The accuracy of the calculated eigenvectors improves as dgmr
c proceeds. When solving for multiple rhs, it usually helps to use 
c eigenvectors calculated in solving for a previous rhs, as the 
c starting eigenvectors when solving for subsequent rhs. To do this, 
c set ip=0 for the first rhs, and dbgmr automatically sets ip=neig 
c for subsequent rhs. Below is an example of the calling 
c sequence to do this.
c
c     neig=positive integer less than im-1, typical values is 4
c     do 100 ii=1,nOfRhs
c       assign rhs
c       assign initial guess sol
c       if(ii.eq.1)ip=0
c       call dbgmr(....)
c         if
c           :
c           : reverse communication code
c           :
c         end if
c 100 continue
c
c In the above example it is assumed that im, and neig are
c the same for each rhs. These can be changed, but
c there are the following restrictions.
c - ip must be less than or equal to the number of eigenvectors
c   stored in vv(1,ifirst), ... vv(1,ifirst+neig-1)
c - eigenvectors are stored in vv(1,ifirst), vv(1,ifirst+1), ...
c   vv(1,ifirst+neig-1), where ifirst=im+2+2. If im is different for 
c   different rhs, copy eigenvectors from locations corresponding to 
c   old value of ifirst to location corresponding to new value of ifirst.
c - when subroutine deigen is called from dgmr, either ip is equal
c   to zero, and neig can take on any value, or ip is greater than
c   zero, and  neig must be less than or equal to ip+2. 
c   This requirement neig<=ip+2 is due to the double use of space 
c   in deigen to save memory. The number of eigenvectors used to 
c   solve for each rhs can thus be increased from one rhs to the 
c   next, but not by more than 2 each time.
c
c-----------------------------------------------------------------------
c Notes on V_{m+1} and W_{m}
c --------------------------
c W and V are subspaces satisfying A M^{-1} W_{m} = V_{m+1} {\bar H}_{m}.
c Here A is the coefficient matrix, M^{-1} is the preconditioning operation,
c and {\bar H}_m is an m+1 x m matrix (Hessenberg + one row). If r_0 is the
c initial residual, and u_1 ... u_{ip} are the ip eigenvectors added in
c deflation, then V is a basis for
c
c span[r_0, Au_1 ... Au_{ip} , Ar_0     ... A_{im-ip-1}r_0, A_{im-ip}r_0]
c and we can write
c V = [v_1, v_2  ... v_{ip+1}, v_{ip+2} ... v_{im},         v_{im+1}]
c
c W is a basis for
c span[u_1 ... u_{ip}, r_0   , Ar_0     ... A_{im-ip-1}r_0]
c and we can write
c W = [u_1 ... u_{ip}, v_1   , v_{ip+2} ... v_{im}]
c
c The vectors v_1 ... v_{im+1} are stored at vv(1,1) ... vv(1,im+1),
c and the vectors u_1 ... u_{ip} are stored at vv(1,ifirst) ...
c vv(1,ifirst+ip-1)
c
c-----------------------------------------------------------------------
c     local variables --
c-----------------------------------------------------------------------
      integer KryDim 
      parameter (KryDim = 200) 
c     Arnoldi size should not exceed KryDim in this version..
c     to reset change parameter statement above 
c-----------------------------------------------------------------------
      real*8  hh(KryDim+1,KryDim),c(KryDim),s(KryDim),rs(KryDim+1)
      real*8  work1(KryDim+1,KryDim),work2(KryDim+1,KryDim)
      real*8  t,ro,pddot,epsmac,eps1,gam, ops 
      integer ifirst,iwork,j,k,i1,k1,ie,label,isave,ierr
c
c hh      - R matrix of QR decomposition of GMRES matrix {\bar H}
c work1,2 - work matrices used in calculation of eigenvectors
c c,s     - contain rotations to calculate R from {\bar H}
c rs      - right hand side of Hessenberg system
c
c     ---work arrays used in deigen
      real*8  alfr(KryDim),alfi(KryDim),beta(KryDim),eigMag(KryDim)
      real*8  eigScale(KryDim)
      integer ieigOrder(KryDim)
c-----------------------------------------------------------------------
      data epsmac/1.d-16/
c-----------------------------------------------------------------------
      save
c
      if (icode .eq. 0) assign 100 to label 
      goto label 
c
 100  continue
      if(nvv.lt.n*(im+1 + max(ip,neig) + 2))then
  	write(*,*)'nvv,n,im,neig = ',nvv,n,im,neig
     	stop 'in dgmr require nvv >= n*(im+1 + max(ip,neig) + 2)'
      end if
      if(im.gt.KryDim)stop 'in dgmr im.gt.KryDim'
      if(ip.ne.0 .and. neig.gt.ip+2)stop  
     ;'in dgmr ip.ne.0 .and. neig.gt.ip+2'
      ifirst=im+2+2
      iwork=im+2
      its = 0
c
c     ===beginning of very first Arnoldi vector===
c
c-----------------------------------------------------------------------
c     **  beginning outer loop. compute initial residual vector
c-----------------------------------------------------------------------
 10   continue
      i = 0 
c
c     calculate initial residual = first vector of Krylov subspace
      call dcopy (n, sol, 1, vv(1,2), 1) 
      ipi   = 1
      ipo   = ipi+n 
      icode = 2 
      assign 11 to label 
      return
 11   continue
      call dscal(n,-1.d0, vv,  1)
      call daxpy(n, 1.d0, rhs, 1, vv, 1)
      ro = pddot(n, vv, 1, vv, 1)
      ro = dsqrt(ro)
      if (iout .gt. 0) print 199, its, ro
      if (abs(ro) .le. eps) goto 999 
      call dscal(n, 1.d0/ro, vv, 1)
      if (its .eq. 0)eps1=eps*ro
c     
c     initialize 1-st term  of rhs of Hessenberg system..
c     
      rs(1) = ro
      i  = 0
      ie = 0
c
c     ===beginning of adding new vector to Krylov subspace===
 4    i   = i   + 1
      i1  = i   + 1
c
c     add next vector
c
c     preconditioning operation
c
      if(i.le.ip)then
  	ie  = ie+1
   	ipo = (ifirst-1+ie-1)*n + 1
      else if(i.eq.ip+1 .and. ip.gt.0)then
  	ipo=1
      else
  	ipo = (i-1)*n + 1
      end if
      ipi = (iwork-1)*n + 1
      icode = 1
      assign 200 to label
      return
 200  continue
c     
c     call matvec operation
c     
      ipo = ipi
      ipi = i*n + 1 
      icode = 2
      assign 300 to label
      return
 300  continue
c     
cc    modified Gram - Schmidt with partial reorthogonalization...
c      call mgsr(n,1,i1,vv,hh(1,i))
      call mgsro(.false.,n,n,im+1,i1,ops,vv,hh(1,i),ierr)
      if(ierr.ne.0) then
         print *, 'ierr = ', ierr, ' in mgsro, exit code'
         stop ' mgsro -- error  '
      endif
cc     
c     done with modified Gram Schmidt and Arnoldi step. 
c     now  update factorization of hh
c     
c     perform previous transformations  on i-th column of h
c     
 58   if (i.gt.1)then
 	do 66 k=2,i
 	  k1 = k-1
 	  t = hh(k1,i)
 	  hh(k1,i) = c(k1)*t + s(k1)*hh(k,i)
 	  hh(k,i) = -s(k1)*t + c(k1)*hh(k,i)
 66	continue
      end if
      gam = dsqrt(hh(i,i)**2 + hh(i1,i)**2)
      if (gam .eq. 0.0d0) gam = epsmac
c
c     determine place rotation
c 
      c(i) = hh(i,i)/gam
      s(i) = hh(i1,i)/gam
      rs(i1) = -s(i)*rs(i)
      rs(i) =  c(i)*rs(i)
c     
c     determine res. norm. and test for convergence-
c     
      hh(i,i) = c(i)*hh(i,i) + s(i)*hh(i1,i)
      its = its + 1
      ro = abs(rs(i1))
*  Changed this to print to screen
      if (iout .gt. 0) print 199, its, ro
c     write(1,"(i6,e12.4)") its, alog10(ro)
c     add another vector to Krylov subspace if no convergence
      if((ro .gt. eps1).and.(i.lt.im).and.(its.lt.maxits))goto 4
c     
c     now compute solution. first solve upper triangular system.
c     
      rs(i) = rs(i)/hh(i,i)
      do 30 k = i-1, 1, -1 
         t=rs(k)
         do 40 j=k+1,i
            t = t-hh(k,j)*rs(j)
 40      continue
         rs(k) = t/hh(k,k)
 30   continue
c     
c     done with back substitution..
c     now form linear combination to get solution
      if(ip.eq.0 .or. i.eq.1)then
        call dgemv('n',n,i,1.d0,vv,n,rs,1,0.d0,vv(1,iwork),1)
      else
  	isave=min(i,ip)
  	call dgemv('n',n,isave,1.d0,vv(1,ifirst),n,rs,1,0.d0,
     .	vv(1,iwork),1)
  	if(i.gt.ip)call dgemv('n',n,1,1.d0,vv,n,rs(ip+1),
     .	1,1.d0,vv(1,iwork),1)
   	if(i.gt.ip+1)call dgemv('n',n,i-(ip+1),1.d0,vv(1,ip+2),
     .	n,rs(ip+2),1,1.d0,vv(1,iwork),1)
      end if
c     
c     ---preconditioning operation
      ipo = (iwork-1)*n+1
      ipi = (iwork  )*n+1
      icode=1
      assign 600 to label
      return
 600  continue
c     ---update solution
      call daxpy(n,1.d0,vv(1,iwork+1),1,sol,1)
c
c     Conditional statement for single rhs. At restart call
c     deigen to calculate eigenvectors
      if ( neig.gt.0 .and. (ro.gt.eps1) ) then
c
c     Conditional statement for multiple rhs. At restart, or at
c     convergence, call routines to calculate eigenvectors
c      if ( neig.gt.0 .and. ((ro.gt.eps1).or.(i.gt.neig)) ) then
c       ---calculate eigenvectors
      	call deigen(n,i,im,ip,neig,work1,work2,hh,vv,ifirst,
     ;	iwork,s,c,KryDim,alfr,alfi,beta,eigMag,eigScale,ieigOrder)
c       ---calculate residuals in eigenvectors/eigenvealues
        icode=0
 500    continue
          call eigResid(n,neig,vv,ifirst,iwork,ioutEig,icode,
     ;	  ipo,ipi,KryDim,alfr,alfi,beta,eigMag,eigScale,ieigOrder)
            if(icode.eq.0) go to 700
  	    if( (icode.ne.1) .and. (icode.ne.2) )
     ;      stop 'in deigen, icode must equal 0, 1, or 2'
  	  assign 400 to label
  	  return
 400	  continue
  	  go to 500
 700    continue
      end if
c
c     test for return 
c     
      if (ro  .le. eps1)   go to 999
      if (its .ge. maxits) go to 998
c
c     restart dgmr
      goto 10
c
c---------------------------------------------------------------
c
 199  format('   -- dgmres its =', i4, ' res. norm =', d20.6)
c     
 999  icode = 0
      return
 998  icode = -1
      return
c-----end-of-dgmr----------------------------------------------------- 
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
      subroutine mgsr (n, i0, i1, ss, r)
      integer n, i0, i1
      real*8 ss(n,*), r(*)
c-----------------------------------------------------------------------
c Modified Gram - Schmidt  with  partial  reortho. The vector ss(*,i1) is
c orthogonalized against vectors ss(*,i0)...ss(*,i1-1)  (which  are  already
c orthogonal).  The coefficients of the orthogonalization are returned in
c the array r
c------------------------------------------------------------------------
      real*8 hinorm, tet, pddot, t, dsqrt
      data  tet/10.0d0/
clbw  do 53 j=i0,i1
      do 53 j=1, i1
         r(j) = 0.0d0
 53   continue
c
c ending position
c
      iend = i1-1
      it = 0
 54   hinorm = 0.0d0
c
c number of outer Gram-Schmidt iterations
c
      it = it +1
      do 55 j=i0, iend
         t = pddot(n, ss(1,j),1,ss(1,i1),1)
         hinorm = hinorm + t**2
         r(j) = r(j) + t
         call daxpy(n,-t,ss(1,j),1,ss(1,i1),1)
 55   continue
      t = pddot(n, ss(1,i1), 1, ss(1,i1), 1)
c
c     test for reorthogonalization see daniel et. al.
c     two reorthogonalization allowed ---
c
      if (t*tet .le. hinorm .and. it .lt. 2) goto 54
      t =dsqrt(t)
      r(i1)= t
      if (t .eq. 0.0d0) return
      t = 1.0d0/t
      do 57  k=1,n
         ss(k,i1) = ss(k,i1)*t
 57   continue
      return
c----------end-of-mgsr--------------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c
      subroutine deigen(ndim,iii,im,ip,neig,work1,work2,hh,vv,ifirst,
     ;	iwork,s,c,KryDim,alfr,alfi,beta,eigMag,eigScale,ieigOrder)
      implicit none
      integer ndim,iii,im,ip,neig,ifirst,iwork,KryDim,ieigOrder(KryDim)
      real*8  hh(KryDim+1,*),vv(ndim,*),s(KryDim),c(KryDim)
      real*8  work1(KryDim+1,*),work2(KryDim+1,*),eigScale(KryDim)
      real*8  alfr(KryDim),alfi(KryDim),beta(KryDim),eigMag(KryDim)
      save
c-----------------------------------------------------------------------
c ---input arguments
c ndim	  - size of matrix A
c iii	  - number of Arnoldi vectors, including eigenvectors
c im	  - maximum number of Arnoldi vectors, including eigenvectors
c ip      - number of eigenvectors in subspace W when subroutine called
c neig	  - number of eigenvectors to be calculated
c work1,2 - work arrays, must have same dimension as hh
c hh   	  - R matrix of QR decomposition of GMRES matrix {\bar H}_m
c vv      - array used to store Krylov subspace V_{n+1} at vv(1,1),
c           eigenvectors of W{n} at vv(1,ifirst), and workspace at 
c           vv(1,iwork). Here W and V are the subspaces from dgmr
c	    satisfying   A M^{-1} W_{m} = V_{m+1} {\bar H}_{m} 
c           Size of vv must be >= ndim * ( im+1 + max(ip,neig) + 2 )
c ifirst  - points to first column of vv containing eigenvectors
c iwork   - points to first column of work space of vv
c s,c     - contain rotations for QR decomposition of {\bar H}_m
c KryDim  - number of vectors in Krylov subspace, and leading dimension
c	    of arrays
c
c ---output arguments
c vv      - contains neig eigenvectors in vv(1,ifirst),vv(1,ifirst+1),...
c	    vv(1,ifirst+neig-1)
c
c ---work arrays
c alfr     - denominator of real part of eigenvalue
c alfi     - denominator of imaginary part of eigenvalue
c beta     - numerator of eigenvalue
c eigMag   - magnitude of eigenvalue
c eigScale - scaling for eigenvectors
c ieigOrder- order of eigenvalues
c-----------------------------------------------------------------------
c ---Description---
c deigen is to be used with dgmr. It calls eispack subroutine rgg 
c to solve the generalized eigenvalue problem                              
c                                         
c	 F z  = 1/q R z        where   F = [_  Q_m V^{H}_{m+1} W_m _]
c
c                                      here [_  _] denotes that the last
c				       row of the matrix is deleted
c					   
c
c The eigen vectors are transformed using
c        y  = W z 
c Eigenvectors y are ordered from smallest to largest
c   
c Here W_m,V_{m+1},{\bar H}_m are matrices from dgmr given by
c              A M^{-1} W_m = V_{m+1} {\bar H}_m
c A is the coefficient matrix, M^{-1} is the preconditioning operation,
c and {\bar H}_m is an m+1 x m matrix (Hessenberg + one row). If r_0 is the
c initial residual, and u_1 ... u_{ip} are the ip eigenvectors added in 
c deflation, then V is a basis for 
c
c span[r_0, Au_1 ... Au_{ip} , Ar_0     ... A_{im-ip-1}r_0, A_{im-ip}r_0]
c and we can write
c V = [v_1, v_2  ... v_{ip+1}, v_{ip+2} ... v_{im},         v_{im+1}]
c
c W is a basis for 
c span[u_1 ... u_{ip}, r_0   , Ar_0     ... A_{im-ip-1}r_0]
c and we can write
c W = [u_1 ... u_{ip}, v_1   , v_{ip+2} ... v_{im}]
c
c R and Q are matrices of the QR decomposition of {\bar H}_m
c		Q {\bar H}_m = R
c
c If ndim is the dimension of A, and m is the number of Arnoldi vectors, 
c including eigenvectors, then the arrays have the following nonzero entries
c  {\bar H}_m(m+1, m)
c           R(m,   m)
c 	    Q(m+1, m)
c           A(ndim,ndim)
c         W_m(ndim,m)
c     V_{m+1}(ndim,iii+1)
c References:
c -R B Morgan 'A restarted GMRES method augmented with eigenvectors'
c  SIAM J. Matrix Analysis and Applications , 16(4):1154-1171,1995.
c -Andrew Chapman and Yousef Saad 'Deflated and augmented Krylov
c  subspace techniques' MSI report UMSI 95/181 September 1995
c-----------------------------------------------------------------------
c     ---local variable declarations---
      real*8  pddot,save1,save2,error,denom
      integer ior,imatz,ii,jj,kk,ierr,labelE,ldh,isave,isave2
c
c-----------------------------------------------------------------------
c
      ldh=KryDim+1
      if(neig.eq.0) return
      if(iii.gt.ldh)stop    'in deigen iii > ldh'
      if(iii.gt.KryDim)stop 'in deigen iii > dDim' 
      if(neig.gt.iii)then
  	write(*,"('number eigenvectors: requested =',2I6)")neig
  	write(*,"('number eigenvectors: available =',2I6)")iii
  	write(*,"('*** reduce number eigenvectors requested***')")
  	stop 'area 3 ' 
      end if
c
c-------------------------------------------------------------------
c     ---calculate V^H W
      if(ip.eq.0)then
        do jj=1,iii
          do ii=1,iii+1
            work1(ii,jj)=0.
          end do
          work1(jj,jj)=1.
        end do
      else
  	isave=min(iii,ip)
        call dgemm('t','n',iii+1,isave,ndim,1.d0,
     ;  vv,ndim,vv(1,ifirst),ndim,0.d0,work1,ldh)
        do jj=ip+1,iii
          do ii=1,iii+1
            work1(ii,jj)=0.
          end do
        end do
  	work1(1,ip+1)=1.
  	do jj=ip+2,iii
  	  work1(jj,jj)=1.
  	end do
      end if
cc
c     ---calculate Q V^H W by applying  rotations  to work1
      do ii=1,iii
c       perform previous transformations  on i-th column of h
        do 66 kk=2,iii+1
          jj = kk-1
          save1= work1(jj,ii)
          work1(jj,ii) =  c(jj)*save1 + s(jj)*work1(kk,ii)
          work1(kk,ii) = -s(jj)*save1 + c(jj)*work1(kk,ii)
 66     continue
      end do
c
c     ---zero elements of hh below diagonal
      do jj=1,iii
  	do ii=jj+1,iii+1
  	  hh(ii,jj)=0.
  	end do
      end do
c
c--------------------------------------------------------------------
c     --- call eispack subroutine to calculate 'z' vectors
      imatz=1
      call rgg(KryDim+1,iii,work1,hh,alfr,alfi,beta,imatz,work2,ierr)
      if(ierr.ne.0)then
 	write(*,*)'*** error in rgg, more than 30 * ',KryDim,' ***'
 	write(*,*)'iteraitons used in trying to calculate eigenvalue'
 	stop 'area  4 ' 
      end if
c
c--------------------------------------------------------------------
c     --- calculate order of eigenvectors y so that their corresponding
c     --- eigenvalues are in order from smallest to largest. This 
c     --- corresponds to largest to smallest for 'z' vectors 
c     --- (stored in work2)
      call norder(iii,eigMag,ieigOrder,alfr,alfi,beta,
     ;	work2,KryDim+1)
c
c--------------------------------------------------------------------
c     ---scaling 'z' vectors
      do ii=1,iii
        save1 = pddot(iii, work2(1,ii), 1, work2(1,ii), 1)
        save1 = 1.0 / dsqrt(save1)
        call dscal(iii,save1,work2(1,ii),1)
    	eigScale(ii)=save1
      end do
c
c--------------------------------------------------------------------
c     --- Transforming 'z' vectors to 'y = W z' eigenvectors ---
c     --- y eigenvectors stored in vv(1,ifirst)
      if(ip.eq.0)then
  	call dgemm('n','n',ndim,neig,iii,1.d0,vv,ndim,work2,ldh,
     ;  0.d0,vv(1,ifirst),ndim)
      else
c       note that eigenvectors are stored temporarily in 
c       vv(1,2)...vv(1,ip+1),vv(1,iwork),vv(1,iwork+1)
cc      This requires that neig.le.ip+2
        if(neig.gt.ip+2)then
          write(*,*)'*************************************************'
          write(*,*)'*  in subroutine deigen, require neig<=ip+2     *'
          write(*,*)'*  This is for memory requirements. If code     *'
          write(*,*)'*  needs neig>ip+2 , increase size of vv and    *'
          write(*,*)'*  value of ifirst to allow for more work space *'
          write(*,*)'*************************************************'
          stop 'area 5 ' 
        end if
  	isave2=min(neig,ip)
  	isave=min(iii,ip)
        call dgemm('n','n',ndim,isave2,isave,1.d0,vv(1,ifirst),ndim,
     ;  work2,ldh,0.d0,vv(1,2),ndim)
        if(iii.gt.ip)
     ;  call dgemm('n','n',ndim,isave2,1,1.d0,vv,ndim,
     ;  work2(ip+1,1),ldh,1.d0,vv(1,2),ndim)
        if(iii.gt.ip+1)
     ;  call dgemm('n','n',ndim,isave2,iii-ip-1,1.d0,vv(1,ip+2),
     ;  ndim,work2(ip+2,1),ldh,1.d0,vv(1,2),ndim)
c       ---copy eigenvectors into vv(1,ifirst), vv(1,ifirst+1), ...
    	call dcopy(ndim*isave2,vv(1,2),1,vv(1,ifirst),1)
c
        if(neig.gt.ip)then
  	  isave2=neig-ip
  	  isave =min(iii,ip)
          call dgemm('n','n',ndim,isave2,isave,1.d0,vv(1,ifirst),ndim,
     ;    work2(1,ip+1),ldh,0.d0,vv(1,iwork),ndim)
          if(iii.gt.ip)
     ;    call dgemm('n','n',ndim,isave2,1,1.d0,vv,ndim,
     ;    work2(ip+1,ip+1),ldh,1.d0,vv(1,iwork),ndim)
          if(iii.gt.ip+1)
     ;    call dgemm('n','n',ndim,isave2,iii-ip-1,1.d0,vv(1,ip+2),
     ;    ndim,work2(ip+2,ip+1),ldh,1.d0,vv(1,iwork),ndim)
c         ---copy eigenvect into vv(1,ifirst+ip), vv(1,ifirst+ip+1), ...
          call dcopy(ndim*isave2,vv(1,iwork),1,vv(1,ifirst+ip),1)
        end if
      end if
c
c     ---set ip equal to the number of eigenvectors in the subspace
      ip=neig
c
c--------------------------------------------------------------------
c     ---scaling eigenvectors
      do ii=1,neig
        save1 = pddot(ndim, vv(1,ifirst-1+ii), 1, vv(1,ifirst-1+ii),1)
        save1 = 1.0 / dsqrt(save1)
        call dscal(ndim,save1,vv(1,ifirst-1+ii),1)
    	eigScale(ii)=eigScale(ii)*save1
      end do
c
      return
c----------end-of-deigen------------------------------------------------
c-----------------------------------------------------------------------
      end
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
      subroutine eigResid(ndim,neig,vv,ifirst,iwork,ioutEig,icodeE,
     ;	ipo,ipi,KryDim,alfr,alfi,beta,eigMag,eigScale,ieigOrder)
      implicit none
      integer ndim,neig,ifirst,iwork,icodeE,ipo,ipi,ioutEig,KryDim
      integer ieigOrder(KryDim)
      real*8  vv(ndim,*),alfr(KryDim),alfi(KryDim),beta(KryDim)
      real*8  eigMag(KryDim),eigScale(KryDim)
      save
c-----------------------------------------------------------------------
c ---input arguments
c ndim	   - size of matrix A
c neig	   - number of eigenvectors
c vv       - contains neig eigenvectors in vv(1,ifirst),vv(1,ifirst+1),...
c ifirst   - points to first column of vv containing eigenvectors
c iwork    - points to first column of work space of vv
c ioutEig  - iounit number for printing eigenvalue/vector residual. if
c            ioutEig = 0, no eigenvalue/vector residuals are calculated
c            or printed. This file needs to be opened in the calling
c            program
c icodeE   - flag for reverse communication
c ipo,ipi  - pointers for reverse communication
c KryDim   - dimension of Krylov subspace, and leading dimension of
c	     arrays containing eigenvector information
c alfr     - denominator of real part of eigenvalue
c alfi     - denominator of imaginary part of eigenvalue
c beta     - numerator of eigenvalue
c eigMag   - magnitude of eigenvalue
c eigScale - scaling for eigenvectors
c ieigOrder- order of eigenvalues
c
c ---output arguments
c none
c
c-----------------------------------------------------------------------
c ---Description---
c eigResid is to be used with subroutine deigen. It is used to calculate
c the residual in the estimated eigenvector/values. Note that calculating 
c the residuals is time consuming, and they are they are calculated only 
c to be written to the file ioutEig, they are never used. Set ioutEig=0, 
c unless you are interested in seeing the residual in the calculated
c eigenvectors.
c-----------------------------------------------------------------------
c REVERSE COMMUNICATION
c----------------------
c USAGE: eigResid should be put in a loop and the loop should be
c active for as long as icode is not equal to 0. On return dgmr
c will ask the user to perform an operation on vv(ipo), and put the
c result in vv(ipi). The operation could be a matrix vector 
c multiplication, or a preconditioning operation. After the 
c operation, call eigResid again. This should be done for as long as 
c icode .ne. 0.  On entry icode should always be equal to 0.
c * if icode  =  0 --> terminate loop
c * if icode  =  1 --> operation requested is preconditioning
c * if icode  =  2 --> operation requested is a matrix-vector product.
c-----------------------------------------------------------------------
c     ---local variable declarations---
      real*8  pddot,save1,save2,error,denom
      integer ior,ii,isave,isave2,labelE
c
c-----------------------------------------------------------------------
      if(icodeE.eq.0)assign 100 to labelE
      go to labelE
 100  continue
c
c--------------------------------------------------------------------
c     --- Calculate residual || A yy - alpha yy || for estimated
c     --- eigenvalues and eigenvectors
c
      if(ioutEig.eq.0)return
cc      write(ioutEig,*)
cc      write(ioutEig,*)
cc     ;'                        ________           ____       ___'
cc      write(ioutEig,*)
cc     ;' error=dsqrt((Ayy-a yy , Ayy-a yy)/((a yy , a yy)+(Ayy,Ayy))'
cc      write(ioutEig,*)' ior        eigenvalue         error'
cc      write(ioutEig,*)'         real        imag'
      ior=1
 300  continue
      if(alfi(ior).eq.0)then
c  	---eigenvalue and eigenvector are real
cc      ---preconditioning operation---
        if(ior.gt.neig)go to 400
        call dcopy(ndim,vv(1,ifirst-1+ior),1,vv(1,iwork),1)
  	ipo=(iwork-1)*ndim+1
  	ipi=(iwork  )*ndim+1
  	icodeE=1
  	assign 2000 to labelE
  	return
 2000	continue
cc      ---matrix vector multiplication---
  	ipo = ipi
  	ipi = (iwork-1)*ndim+1
  	icodeE = 2
  	assign 5000 to labelE
  	return
 5000	continue
c
   	call dcopy(ndim,vv(1,ifirst-1+ior),1,vv(1,iwork+1),1)
   	call dscal(ndim,beta(ior)/alfr(ior),vv(1,iwork+1),1)
      	denom=0.
  	do ii=1,ndim
  	  denom=denom+(abs(vv(ii,iwork))+abs(vv(ii,iwork+1)))*
     ;	              (abs(vv(ii,iwork))+abs(vv(ii,iwork+1)))
   	end do
  	denom=dsqrt(denom)
        call daxpy(ndim,-1.d0,vv(1,iwork+1),1,vv(1,iwork),1)
        error = pddot(ndim, vv(1,iwork),1,vv(1,iwork),1)
        error = dsqrt(error) / denom
cc 	write(ioutEig,"(I4,4e12.4)")ior,beta(ior)/alfr(ior),0.,error
 	ior=ior+1
      else
cc 	---eigenvalue and eigenvector are complex
cc
        if(ior.gt.neig-1)then
          save2=beta(ior)/(alfr(ior)*alfr(ior)+alfi(ior)*alfi(ior))
          save1= alfr(ior)*save2
          save2=-alfi(ior)*save2
          write(ioutEig,"(I4,2e12.4,'complex part not available')")
     ;    ior,save1,save2
          go to 400
        else
cc        remove scaling by multiplying complex part
          save1=eigScale(ior)/eigScale(ior+1)
          call dscal(ndim,save1,vv(1,ifirst+ior),1)
          save2=beta(ior)/(alfr(ior)*alfr(ior)+alfi(ior)*alfi(ior))
          save1= alfr(ior)*save2
          save2=-alfi(ior)*save2
        end if
cc
cc      ---preconditioning operation---
        call dcopy(ndim,vv(1,ifirst-1+ior),1,vv(1,iwork),1)
  	ipo=(iwork-1)*ndim+1
  	ipi=(iwork  )*ndim+1
  	icodeE=1
  	assign 3000 to labelE
  	return
 3000	continue
cc      ---matrix vector multiplication---
        ipo = ipi
        ipi = (iwork-1)*ndim+1
        icodeE = 2
        assign 6000 to labelE
        return
 6000   continue
cc
   	call dcopy(ndim,vv(1,ifirst-1+ior),  1,vv(1,iwork+1),1)
  	call dscal(ndim,save1,vv(1,iwork+1),1)
        call daxpy(ndim,-save2,vv(1,ifirst+ior),1,vv(1,iwork+1),1)
        denom=0.
        do ii=1,ndim
          denom=denom+(abs(vv(ii,iwork))+abs(vv(1,iwork+1)))*
     ;                (abs(vv(ii,iwork))+abs(vv(1,iwork+1)))
        end do
        call daxpy(ndim,-1.d0,vv(1,iwork+1),1,vv(1,iwork),1)
  	error=pddot(ndim,vv(1,iwork),1,vv(1,iwork),1)
cc
cc
cc      ---preconditioning operation---
        call dcopy(ndim,vv(1,ifirst+ior),1,vv(1,iwork),1)
  	ipo=(iwork-1)*ndim+1
  	ipi=(iwork  )*ndim+1
  	icodeE=1
  	assign 4000 to labelE
  	return
 4000	continue
cc      ---matrix vector multiplication---
        ipo = ipi
        ipi = (iwork-1)*ndim+1
        icodeE = 2
        assign 7000 to labelE
        return
 7000   continue
cc
    	call dcopy(ndim,vv(1,ifirst+ior),1,vv(1,iwork+1),1)
  	call dscal(ndim,save1,vv(1,iwork+1),1)
        call daxpy(ndim,save2,vv(1,ifirst-1+ior),1,vv(1,iwork+1),1)
        do ii=1,ndim
          denom=denom+(abs(vv(ii,iwork))+abs(vv(ii,iwork+1)))*
     ;                (abs(vv(ii,iwork))+abs(vv(ii,iwork+1)))
        end do
        call daxpy(ndim,-1.d0,vv(1,iwork+1),1,vv(1,iwork),1)
  	error = error+pddot(ndim,vv(1,iwork),1,vv(1,iwork),1)
  	error = dsqrt(error/denom)
cc 	write(ioutEig,"(I4,4e12.4)")ior,save1,save2,error
c
c  	replace scaling by multiplying complex part
  	save1=eigScale(ior+1)/eigScale(ior)
   	call dscal(ndim,save1,vv(1,ifirst+ior),1)
c
        ior=ior+2
      end if
      go to 300
 400  icodeE=0
c
cc      write(ioutEig,*)'==============================================='
cc      write(ioutEig,*)'==============================================='
cc      write(ioutEig,*)'==============================================='
c
      return
c-----------end-of-eigResid-----------------------------------------
c-------------------------------------------------------------------
      end
c-------------------------------------------------------------------
c--------------------------------------------------------------------
      subroutine norder(iii,eigMag,ieigOrder,alfr,alfi,beta,
     ;  work,ldw)
      implicit none
      integer ieigOrder(*),iii,ldw
      real*8  eigMag(*),alfr(*),alfi(*),beta(*),work(ldw,*)
      integer ii,jj,imax
      real*8  rmax,rmag,tmp,epsmac
      data epsmac /1.d-16/
c
c subroutine orders eigenvalues and eigenvectors from eispack subroutine
c rgg so that eigenvalues are from largest to smallest.
c
c ---entry---
c iii           number of eigenvalues calculated by rgg
c alfr,alfi     real and imag parts of eigenvalues
c beta          denominator of eigenvalues
c work          array containing eigenvectors in columns
c ldw           leading dimension of work
c
c ---on return---
c alfr,alfi,beta,work   eigenvalues and vectors, ordered largest
c                       to smallest
c
c ---work---
c eigMag        work array, contains magnitudes of eigenvalues
c ieigOrder     work array, keeps order of eigenvalues
c
c     calculate magnitude and location of eigenvalues
      do ii=1,iii           
            tmp = alfr(ii)*alfr(ii) + alfi(ii)*alfi(ii) 
            tmp = tmp / (tmp + beta(ii)*beta(ii) + epsmac)
            eigMag(ii)=  tmp  
            ieigOrder(ii)=ii
      end do
c
c     calculate order of eigenvalues from largest to smallest
      do ii=1,iii-1
        imax=ii
        rmax=eigMag(ieigOrder(ii))
        do jj=ii+1,iii
          rmag=eigMag(ieigOrder(jj))
          if(rmag.gt.rmax) then
            imax=jj
            rmax=rmag
          end if
        end do
        jj=ieigOrder(ii)
        ieigOrder(ii)  =ieigOrder(imax)
        ieigOrder(imax)=jj
      end do
c
c     make sure that for complex eigenvalues 
c     (which occur as d so that the real part of
c     the eigenvectors occurs first
      ii=1
 100  if(ii.ge.iii)go to 200
        if(alfi(ieigOrder(ii)).eq.0.)then
          ii=ii+1
        else
          if(alfi(ieigOrder(ii)).lt.0.)then
            jj=ieigOrder(ii+1)
            ieigOrder(ii+1)=ieigOrder(ii)
            ieigOrder(ii)=jj
          end if
          ii=ii+2
        end if
      go to 100
 200  continue
c
c     reorder eigenvalues and eigenvectors
      do ii=1,iii
        jj=ieigOrder(ii)
        tmp=alfr(jj)
        alfr(jj)=alfr(ii)
        alfr(ii)=tmp
        tmp=alfi(jj)
        alfi(jj)=alfi(ii)
        alfi(ii)=tmp
        tmp=beta(jj)
        beta(jj)=beta(ii)
        beta(ii)=tmp
        call dswap(ldw,work(1,jj),1,work(1,ii),1)
        do jj=ii+1,iii
          if(ieigOrder(jj).eq.ii)ieigOrder(jj)=ieigOrder(ii)
        end do
      end do
c
c------------------------------------------------------------
c FOR FRINT OF EIGENVALUES MAGNITUDES
c      write(*, 456)  (alfr(ii)/beta(ii), alfi(ii)/beta(ii), 
c     $                ii= 1,iii)
c 456  format(4(e10.3, '+', e10.3, 'i  '))
c------------------------------------------------------------
      return
      end
c
c

      subroutine rgg(nm,n,a,b,alfr,alfi,beta,matz,z,ierr)
c
      integer n,nm,ierr,matz
      double precision a(nm,n),b(nm,n),alfr(n),alfi(n),beta(n),z(nm,n)
      logical tf
c
c     this subroutine calls the recommended sequence of
c     subroutines from the eigensystem subroutine package (eispack)
c     to find the eigenvalues and eigenvectors (if desired)
c     for the real general generalized eigenproblem  ax = (lambda)bx.
c
c     on input
c
c        nm  must be set to the row dimension of the two-dimensional
c        array parameters as declared in the calling program
c        dimension statement.
c
c        n  is the order of the matrices  a  and  b.
c
c        a  contains a real general matrix.
c
c        b  contains a real general matrix.
c
c        matz  is an integer variable set equal to zero if
c        only eigenvalues are desired.  otherwise it is set to
c        any non-zero integer for both eigenvalues and eigenvectors.
c
c     on output
c
c        alfr  and  alfi  contain the real and imaginary parts,
c        respectively, of the numerators of the eigenvalues.
c
c        beta  contains the denominators of the eigenvalues,
c        which are thus given by the ratios  (alfr+i*alfi)/beta.
c        complex conjugate pairs of eigenvalues appear consecutively
c        with the eigenvalue having the positive imaginary part first.
c
c        z  contains the real and imaginary parts of the eigenvectors
c        if matz is not zero.  if the j-th eigenvalue is real, the
c        j-th column of  z  contains its eigenvector.  if the j-th
c        eigenvalue is complex with positive imaginary part, the
c        j-th and (j+1)-th columns of  z  contain the real and
c        imaginary parts of its eigenvector.  the conjugate of this
c        vector is the eigenvector for the conjugate eigenvalue.
c
c        ierr  is an integer output variable set equal to an error
c           completion code described in the documentation for qzit.
c           the normal completion code is zero.
c
c     questions and comments should be directed to burton s. garbow,
c     mathematics and computer science div, argonne national laboratory
c
c     this version dated august 1983.
c
c     ------------------------------------------------------------------
c
      if (n .le. nm) go to 10
      ierr = 10 * n
      go to 50
c
   10 if (matz .ne. 0) go to 20
c     .......... find eigenvalues only ..........
      tf = .false.
      call  qzhes(nm,n,a,b,tf,z)
      call  qzit(nm,n,a,b,0.0d0,tf,z,ierr)
      call  qzval(nm,n,a,b,alfr,alfi,beta,tf,z)
      go to 50
c     .......... find both eigenvalues and eigenvectors ..........
   20 tf = .true.
      call  qzhes(nm,n,a,b,tf,z)
      call  qzit(nm,n,a,b,0.0d0,tf,z,ierr)
      call  qzval(nm,n,a,b,alfr,alfi,beta,tf,z)
      if (ierr .ne. 0) go to 50
      call  qzvec(nm,n,a,b,alfr,alfi,beta,z)
   50 return
      end

      double precision function epslon (x)
      double precision x
c
c     estimate unit roundoff in quantities of size x.
c
      double precision a,b,c,eps
c
c     this program should function properly on all systems
c     satisfying the following two assumptions,
c        1.  the base used in representing floating point
c            numbers is not a power of three.
c        2.  the quantity  a  in statement 10 is represented to 
c            the accuracy used in floating point variables
c            that are stored in memory.
c     the statement number 10 and the go to 10 are intended to
c     force optimizing compilers to generate code satisfying 
c     assumption 2.
c     under these assumptions, it should be true that,
c            a  is not exactly equal to four-thirds,
c            b  has a zero for its last bit or digit,
c            c  is not exactly equal to one,
c            eps  measures the separation of 1.0 from
c                 the next larger floating point number.
c     the developers of eispack would appreciate being informed
c     about any systems where these assumptions do not hold.
c
c     this version dated 4/6/83.
c
      a = 4.0d0/3.0d0
   10 b = a - 1.0d0
      c = b + b + b
      eps = dabs(c-1.0d0)
      if (eps .eq. 0.0d0) go to 10
      epslon = eps*dabs(x)
      return
      end
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
      subroutine mgsro(full,lda,n,m,ind,ops,vec,hh,ierr)
      implicit none
      logical full
      integer lda,m,n,ind,ierr
      real*8  ops,hh(m),vec(lda,m)
c-----------------------------------------------------------------------
c     MGSRO  -- Modified Gram-Schmidt procedure with Selective Re-
c               Orthogonalization
c     The ind'th vector of VEC is orthogonalized against the rest of
c     the vectors.
c
c     The test for performing re-orthogonalization is performed for
c     each indivadual vectors. If the cosine between the two vectors
c     is greater than 0.99 (REORTH = 0.99**2), re-orthogonalization is
c     performed. The norm of the 'new' vector is kept in variable NRM0,
c     and updated after operating with each vector.
c
c     full   -- .ture. if it is necessary to orthogonalize the ind'th
c               against all the vectors vec(:,1:ind-1), vec(:,ind+2:m)
c               .false. only orthogonalize againt vec(:,1:ind-1)
c     lda    -- the leading dimension of VEC
c     n      -- length of the vector in VEC
c     m      -- number of vectors can be stored in VEC
c     ind    -- index to the vector to be changed
c     ops    -- operation counts
c     vec    -- vector of LDA X M storing the vectors
c     hh     -- coefficient of the orthogonalization
c     ierr   -- error code
c               0 : successful return
c               -1: zero input vector
c               -2: input vector contains abnormal numbers
c               -3: input vector is a linear combination of others
c
c     External routines used: real*8 pddot
c-----------------------------------------------------------------------
      integer i,k
      real*8  nrm0, nrm1, fct, thr, pddot, zero, one, reorth
      parameter (zero=0.0D0, one=1.0D0, reorth=0.98D0)
      external pddot
c
c     compute the norm of the input vector
c


      nrm0 = pddot(n,vec(1,ind),1,vec(1,ind),1)
      ops = ops + n + n
      thr = nrm0 * reorth
      if (nrm0.le.zero) then
         ierr = - 1
         return
      else if (nrm0.gt.zero .and. one/nrm0.gt.zero) then
         ierr = 0
      else
         ierr = -2
         return
      endif
      ierr = 0 
c
c     Modified Gram-Schmidt loop
c
      if (full) then
         do 40 i = ind+1, m
            fct = pddot(n,vec(1,ind),1,vec(1,i),1)
            hh(i) = fct
            do 20 k = 1, n
               vec(k,ind) = vec(k,ind) - fct * vec(k,i)
 20         continue
            ops = ops + 4 * n + 2
            if (fct*fct.gt.thr) then
               fct = pddot(n,vec(1,ind),1,vec(1,i),1)
               hh(i) = hh(i) + fct
               do 30 k = 1, n
                  vec(k,ind) = vec(k,ind) - fct * vec(k,i)
 30            continue
               ops = ops + 4*n + 1
            endif
            nrm0 = nrm0 - hh(i) * hh(i)
            if (nrm0.lt.zero) nrm0 = zero
            thr = nrm0 * reorth
 40      continue
      endif
c
      do 70 i = 1, ind-1
         fct = pddot(n,vec(1,ind),1,vec(1,i),1)
         hh(i) = fct
         do 50 k = 1, n
            vec(k,ind) = vec(k,ind) - fct * vec(k,i)
 50      continue
         ops = ops + 4 * n + 2
         if (fct*fct.gt.thr) then
            fct = pddot(n,vec(1,ind),1,vec(1,i),1)
            hh(i) = hh(i) + fct
            do 60 k = 1, n
               vec(k,ind) = vec(k,ind) - fct * vec(k,i)
 60         continue
            ops = ops + 4*n + 1
         endif
         nrm0 = nrm0 - hh(i) * hh(i)
         if (nrm0.lt.zero) nrm0 = zero
         thr = nrm0 * reorth
 70   continue
c
c     test the resulting vector
c
      nrm1 = dsqrt(pddot(n,vec(1,ind),1,vec(1,ind),1))
      ops = ops + n + n
 75   hh(ind) = nrm1
      if (nrm1.le.zero) then
         ierr = -3
         return
      endif

c
c     scale the resulting vector
c
      fct = one / nrm1
      do 80 k = 1, n
         vec(k,ind) = vec(k,ind) * fct
 80   continue
      ops = ops + n + 1
c
c     normal return
c
      ierr = 0
      return
c     end surbotine mgsro
      end


      subroutine qzhes(nm,n,a,b,matz,z)
c
      integer i,j,k,l,n,lb,l1,nm,nk1,nm1,nm2
      double precision a(nm,n),b(nm,n),z(nm,n)
      double precision r,s,t,u1,u2,v1,v2,rho
      logical matz
c
c     this subroutine is the first step of the qz algorithm
c     for solving generalized matrix eigenvalue problems,
c     siam j. numer. anal. 10, 241-256(1973) by moler and stewart.
c
c     this subroutine accepts a pair of real general matrices and
c     reduces one of them to upper hessenberg form and the other
c     to upper triangular form using orthogonal transformations.
c     it is usually followed by  qzit,  qzval  and, possibly,  qzvec.
c
c     on input
c
c        nm must be set to the row dimension of two-dimensional
c          array parameters as declared in the calling program
c          dimension statement.
c
c        n is the order of the matrices.
c
c        a contains a real general matrix.
c
c        b contains a real general matrix.
c
c        matz should be set to .true. if the right hand transformations
c          are to be accumulated for later use in computing
c          eigenvectors, and to .false. otherwise.
c
c     on output
c
c        a has been reduced to upper hessenberg form.  the elements
c          below the first subdiagonal have been set to zero.
c
c        b has been reduced to upper triangular form.  the elements
c          below the main diagonal have been set to zero.
c
c        z contains the product of the right hand transformations if
c          matz has been set to .true.  otherwise, z is not referenced.
c
c     questions and comments should be directed to burton s. garbow,
c     mathematics and computer science div, argonne national laboratory
c
c     this version dated august 1983.
c
c     ------------------------------------------------------------------
c
c     .......... initialize z ..........
      if (.not. matz) go to 10
c
      do 3 j = 1, n
c
         do 2 i = 1, n
            z(i,j) = 0.0d0
    2    continue
c
         z(j,j) = 1.0d0
    3 continue
c     .......... reduce b to upper triangular form ..........
   10 if (n .le. 1) go to 170
      nm1 = n - 1
c
      do 100 l = 1, nm1
         l1 = l + 1
         s = 0.0d0
c
         do 20 i = l1, n
            s = s + dabs(b(i,l))
   20    continue
c
         if (s .eq. 0.0d0) go to 100
         s = s + dabs(b(l,l))
         r = 0.0d0
c
         do 25 i = l, n
            b(i,l) = b(i,l) / s
            r = r + b(i,l)**2
   25    continue
c
         r = dsign(dsqrt(r),b(l,l))
         b(l,l) = b(l,l) + r
         rho = r * b(l,l)
c
         do 50 j = l1, n
            t = 0.0d0
c
            do 30 i = l, n
               t = t + b(i,l) * b(i,j)
   30       continue
c
            t = -t / rho
c
            do 40 i = l, n
               b(i,j) = b(i,j) + t * b(i,l)
   40       continue
c
   50    continue
c
         do 80 j = 1, n
            t = 0.0d0
c
            do 60 i = l, n
               t = t + b(i,l) * a(i,j)
   60       continue
c
            t = -t / rho
c
            do 70 i = l, n
               a(i,j) = a(i,j) + t * b(i,l)
   70       continue
c
   80    continue
c
         b(l,l) = -s * r
c
         do 90 i = l1, n
            b(i,l) = 0.0d0
   90    continue
c
  100 continue
c     .......... reduce a to upper hessenberg form, while
c                keeping b triangular ..........
      if (n .eq. 2) go to 170
      nm2 = n - 2
c
      do 160 k = 1, nm2
         nk1 = nm1 - k
c     .......... for l=n-1 step -1 until k+1 do -- ..........
         do 150 lb = 1, nk1
            l = n - lb
            l1 = l + 1
c     .......... zero a(l+1,k) ..........
            s = dabs(a(l,k)) + dabs(a(l1,k))
            if (s .eq. 0.0d0) go to 150
            u1 = a(l,k) / s
            u2 = a(l1,k) / s
            r = dsign(dsqrt(u1*u1+u2*u2),u1)
            v1 =  -(u1 + r) / r
            v2 = -u2 / r
            u2 = v2 / v1
c
            do 110 j = k, n
               t = a(l,j) + u2 * a(l1,j)
               a(l,j) = a(l,j) + t * v1
               a(l1,j) = a(l1,j) + t * v2
  110       continue
c
            a(l1,k) = 0.0d0
c
            do 120 j = l, n
               t = b(l,j) + u2 * b(l1,j)
               b(l,j) = b(l,j) + t * v1
               b(l1,j) = b(l1,j) + t * v2
  120       continue
c     .......... zero b(l+1,l) ..........
            s = dabs(b(l1,l1)) + dabs(b(l1,l))
            if (s .eq. 0.0d0) go to 150
            u1 = b(l1,l1) / s
            u2 = b(l1,l) / s
            r = dsign(dsqrt(u1*u1+u2*u2),u1)
            v1 =  -(u1 + r) / r
            v2 = -u2 / r
            u2 = v2 / v1
c
            do 130 i = 1, l1
               t = b(i,l1) + u2 * b(i,l)
               b(i,l1) = b(i,l1) + t * v1
               b(i,l) = b(i,l) + t * v2
  130       continue
c
            b(l1,l) = 0.0d0
c
            do 140 i = 1, n
               t = a(i,l1) + u2 * a(i,l)
               a(i,l1) = a(i,l1) + t * v1
               a(i,l) = a(i,l) + t * v2
  140       continue
c
            if (.not. matz) go to 150
c
            do 145 i = 1, n
               t = z(i,l1) + u2 * z(i,l)
               z(i,l1) = z(i,l1) + t * v1
               z(i,l) = z(i,l) + t * v2
  145       continue
c
  150    continue
c
  160 continue
c
  170 return
      end

      subroutine qzit(nm,n,a,b,eps1,matz,z,ierr)
c
      integer i,j,k,l,n,en,k1,k2,ld,ll,l1,na,nm,ish,itn,its,km1,lm1,
     x        enm2,ierr,lor1,enorn
      double precision a(nm,n),b(nm,n),z(nm,n)
      double precision r,s,t,a1,a2,a3,ep,sh,u1,u2,u3,v1,v2,v3,ani,a11,
     x       a12,a21,a22,a33,a34,a43,a44,bni,b11,b12,b22,b33,b34,
     x       b44,epsa,epsb,eps1,anorm,bnorm,epslon
      logical matz,notlas
c
c     this subroutine is the second step of the qz algorithm
c     for solving generalized matrix eigenvalue problems,
c     siam j. numer. anal. 10, 241-256(1973) by moler and stewart,
c     as modified in technical note nasa tn d-7305(1973) by ward.
c
c     this subroutine accepts a pair of real matrices, one of them
c     in upper hessenberg form and the other in upper triangular form.
c     it reduces the hessenberg matrix to quasi-triangular form using
c     orthogonal transformations while maintaining the triangular form
c     of the other matrix.  it is usually preceded by  qzhes  and
c     followed by  qzval  and, possibly,  qzvec.
c
c     on input
c
c        nm must be set to the row dimension of two-dimensional
c          array parameters as declared in the calling program
c          dimension statement.
c
c        n is the order of the matrices.
c
c        a contains a real upper hessenberg matrix.
c
c        b contains a real upper triangular matrix.
c
c        eps1 is a tolerance used to determine negligible elements.
c          eps1 = 0.0 (or negative) may be input, in which case an
c          element will be neglected only if it is less than roundoff
c          error times the norm of its matrix.  if the input eps1 is
c          positive, then an element will be considered negligible
c          if it is less than eps1 times the norm of its matrix.  a
c          positive value of eps1 may result in faster execution,
c          but less accurate results.
c
c        matz should be set to .true. if the right hand transformations
c          are to be accumulated for later use in computing
c          eigenvectors, and to .false. otherwise.
c
c        z contains, if matz has been set to .true., the
c          transformation matrix produced in the reduction
c          by  qzhes, if performed, or else the identity matrix.
c          if matz has been set to .false., z is not referenced.
c
c     on output
c
c        a has been reduced to quasi-triangular form.  the elements
c          below the first subdiagonal are still zero and no two
c          consecutive subdiagonal elements are nonzero.
c
c        b is still in upper triangular form, although its elements
c          have been altered.  the location b(n,1) is used to store
c          eps1 times the norm of b for later use by  qzval  and  qzvec.
c
c        z contains the product of the right hand transformations
c          (for both steps) if matz has been set to .true..
c
c        ierr is set to
c          zero       for normal return,
c          j          if the limit of 30*n iterations is exhausted
c                     while the j-th eigenvalue is being sought.
c
c     questions and comments should be directed to burton s. garbow,
c     mathematics and computer science div, argonne national laboratory
c
c     this version dated august 1983.
c
c     ------------------------------------------------------------------
c
      ierr = 0
c     .......... compute epsa,epsb ..........
      anorm = 0.0d0
      bnorm = 0.0d0
c
      do 30 i = 1, n
         ani = 0.0d0
         if (i .ne. 1) ani = dabs(a(i,i-1))
         bni = 0.0d0
c
         do 20 j = i, n
            ani = ani + dabs(a(i,j))
            bni = bni + dabs(b(i,j))
   20    continue
c
         if (ani .gt. anorm) anorm = ani
         if (bni .gt. bnorm) bnorm = bni
   30 continue
c
      if (anorm .eq. 0.0d0) anorm = 1.0d0
      if (bnorm .eq. 0.0d0) bnorm = 1.0d0
      ep = eps1
      if (ep .gt. 0.0d0) go to 50
c     .......... use roundoff level if eps1 is zero ..........
      ep = epslon(1.0d0)
   50 epsa = ep * anorm
      epsb = ep * bnorm
c     .......... reduce a to quasi-triangular form, while
c                keeping b triangular ..........
      lor1 = 1
      enorn = n
      en = n
      itn = 30*n
c     .......... begin qz step ..........
   60 if (en .le. 2) go to 1001
      if (.not. matz) enorn = en
      its = 0
      na = en - 1
      enm2 = na - 1
   70 ish = 2
c     .......... check for convergence or reducibility.
c                for l=en step -1 until 1 do -- ..........
      do 80 ll = 1, en
         lm1 = en - ll
         l = lm1 + 1
         if (l .eq. 1) go to 95
         if (dabs(a(l,lm1)) .le. epsa) go to 90
   80 continue
c
   90 a(l,lm1) = 0.0d0
      if (l .lt. na) go to 95
c     .......... 1-by-1 or 2-by-2 block isolated ..........
      en = lm1
      go to 60
c     .......... check for small top of b ..........
   95 ld = l
  100 l1 = l + 1
      b11 = b(l,l)
      if (dabs(b11) .gt. epsb) go to 120
      b(l,l) = 0.0d0
      s = dabs(a(l,l)) + dabs(a(l1,l))
      u1 = a(l,l) / s
      u2 = a(l1,l) / s
      r = dsign(dsqrt(u1*u1+u2*u2),u1)
      v1 = -(u1 + r) / r
      v2 = -u2 / r
      u2 = v2 / v1
c
      do 110 j = l, enorn
         t = a(l,j) + u2 * a(l1,j)
         a(l,j) = a(l,j) + t * v1
         a(l1,j) = a(l1,j) + t * v2
         t = b(l,j) + u2 * b(l1,j)
         b(l,j) = b(l,j) + t * v1
         b(l1,j) = b(l1,j) + t * v2
  110 continue
c
      if (l .ne. 1) a(l,lm1) = -a(l,lm1)
      lm1 = l
      l = l1
      go to 90
  120 a11 = a(l,l) / b11
      a21 = a(l1,l) / b11
      if (ish .eq. 1) go to 140
c     .......... iteration strategy ..........
      if (itn .eq. 0) go to 1000
      if (its .eq. 10) go to 155
c     .......... determine type of shift ..........
      b22 = b(l1,l1)
      if (dabs(b22) .lt. epsb) b22 = epsb
      b33 = b(na,na)
      if (dabs(b33) .lt. epsb) b33 = epsb
      b44 = b(en,en)
      if (dabs(b44) .lt. epsb) b44 = epsb
      a33 = a(na,na) / b33
      a34 = a(na,en) / b44
      a43 = a(en,na) / b33
      a44 = a(en,en) / b44
      b34 = b(na,en) / b44
      t = 0.5d0 * (a43 * b34 - a33 - a44)
      r = t * t + a34 * a43 - a33 * a44
      if (r .lt. 0.0d0) go to 150
c     .......... determine single shift zeroth column of a ..........
      ish = 1
      r = dsqrt(r)
      sh = -t + r
      s = -t - r
      if (dabs(s-a44) .lt. dabs(sh-a44)) sh = s
c     .......... look for two consecutive small
c                sub-diagonal elements of a.
c                for l=en-2 step -1 until ld do -- ..........
      do 130 ll = ld, enm2
         l = enm2 + ld - ll
         if (l .eq. ld) go to 140
         lm1 = l - 1
         l1 = l + 1
         t = a(l,l)
         if (dabs(b(l,l)) .gt. epsb) t = t - sh * b(l,l)
         if (dabs(a(l,lm1)) .le. dabs(t/a(l1,l)) * epsa) go to 100
  130 continue
c
  140 a1 = a11 - sh
      a2 = a21
      if (l .ne. ld) a(l,lm1) = -a(l,lm1)
      go to 160
c     .......... determine double shift zeroth column of a ..........
  150 a12 = a(l,l1) / b22
      a22 = a(l1,l1) / b22
      b12 = b(l,l1) / b22
      a1 = ((a33 - a11) * (a44 - a11) - a34 * a43 + a43 * b34 * a11)
     x     / a21 + a12 - a11 * b12
      a2 = (a22 - a11) - a21 * b12 - (a33 - a11) - (a44 - a11)
     x     + a43 * b34
      a3 = a(l1+1,l1) / b22
      go to 160
c     .......... ad hoc shift ..........
  155 a1 = 0.0d0
      a2 = 1.0d0
      a3 = 1.1605d0
  160 its = its + 1
      itn = itn - 1
      if (.not. matz) lor1 = ld
c     .......... main loop ..........
      do 260 k = l, na
         notlas = k .ne. na .and. ish .eq. 2
         k1 = k + 1
         k2 = k + 2
         km1 = max0(k-1,l)
         ll = min0(en,k1+ish)
         if (notlas) go to 190
c     .......... zero a(k+1,k-1) ..........
         if (k .eq. l) go to 170
         a1 = a(k,km1)
         a2 = a(k1,km1)
  170    s = dabs(a1) + dabs(a2)
         if (s .eq. 0.0d0) go to 70
         u1 = a1 / s
         u2 = a2 / s
         r = dsign(dsqrt(u1*u1+u2*u2),u1)
         v1 = -(u1 + r) / r
         v2 = -u2 / r
         u2 = v2 / v1
c
         do 180 j = km1, enorn
            t = a(k,j) + u2 * a(k1,j)
            a(k,j) = a(k,j) + t * v1
            a(k1,j) = a(k1,j) + t * v2
            t = b(k,j) + u2 * b(k1,j)
            b(k,j) = b(k,j) + t * v1
            b(k1,j) = b(k1,j) + t * v2
  180    continue
c
         if (k .ne. l) a(k1,km1) = 0.0d0
         go to 240
c     .......... zero a(k+1,k-1) and a(k+2,k-1) ..........
  190    if (k .eq. l) go to 200
         a1 = a(k,km1)
         a2 = a(k1,km1)
         a3 = a(k2,km1)
  200    s = dabs(a1) + dabs(a2) + dabs(a3)
         if (s .eq. 0.0d0) go to 260
         u1 = a1 / s
         u2 = a2 / s
         u3 = a3 / s
         r = dsign(dsqrt(u1*u1+u2*u2+u3*u3),u1)
         v1 = -(u1 + r) / r
         v2 = -u2 / r
         v3 = -u3 / r
         u2 = v2 / v1
         u3 = v3 / v1
c
         do 210 j = km1, enorn
            t = a(k,j) + u2 * a(k1,j) + u3 * a(k2,j)
            a(k,j) = a(k,j) + t * v1
            a(k1,j) = a(k1,j) + t * v2
            a(k2,j) = a(k2,j) + t * v3
            t = b(k,j) + u2 * b(k1,j) + u3 * b(k2,j)
            b(k,j) = b(k,j) + t * v1
            b(k1,j) = b(k1,j) + t * v2
            b(k2,j) = b(k2,j) + t * v3
  210    continue
c
         if (k .eq. l) go to 220
         a(k1,km1) = 0.0d0
         a(k2,km1) = 0.0d0
c     .......... zero b(k+2,k+1) and b(k+2,k) ..........
  220    s = dabs(b(k2,k2)) + dabs(b(k2,k1)) + dabs(b(k2,k))
         if (s .eq. 0.0d0) go to 240
         u1 = b(k2,k2) / s
         u2 = b(k2,k1) / s
         u3 = b(k2,k) / s
         r = dsign(dsqrt(u1*u1+u2*u2+u3*u3),u1)
         v1 = -(u1 + r) / r
         v2 = -u2 / r
         v3 = -u3 / r
         u2 = v2 / v1
         u3 = v3 / v1
c
         do 230 i = lor1, ll
            t = a(i,k2) + u2 * a(i,k1) + u3 * a(i,k)
            a(i,k2) = a(i,k2) + t * v1
            a(i,k1) = a(i,k1) + t * v2
            a(i,k) = a(i,k) + t * v3
            t = b(i,k2) + u2 * b(i,k1) + u3 * b(i,k)
            b(i,k2) = b(i,k2) + t * v1
            b(i,k1) = b(i,k1) + t * v2
            b(i,k) = b(i,k) + t * v3
  230    continue
c
         b(k2,k) = 0.0d0
         b(k2,k1) = 0.0d0
         if (.not. matz) go to 240
c
         do 235 i = 1, n
            t = z(i,k2) + u2 * z(i,k1) + u3 * z(i,k)
            z(i,k2) = z(i,k2) + t * v1
            z(i,k1) = z(i,k1) + t * v2
            z(i,k) = z(i,k) + t * v3
  235    continue
c     .......... zero b(k+1,k) ..........
  240    s = dabs(b(k1,k1)) + dabs(b(k1,k))
         if (s .eq. 0.0d0) go to 260
         u1 = b(k1,k1) / s
         u2 = b(k1,k) / s
         r = dsign(dsqrt(u1*u1+u2*u2),u1)
         v1 = -(u1 + r) / r
         v2 = -u2 / r
         u2 = v2 / v1
c
         do 250 i = lor1, ll
            t = a(i,k1) + u2 * a(i,k)
            a(i,k1) = a(i,k1) + t * v1
            a(i,k) = a(i,k) + t * v2
            t = b(i,k1) + u2 * b(i,k)
            b(i,k1) = b(i,k1) + t * v1
            b(i,k) = b(i,k) + t * v2
  250    continue
c
         b(k1,k) = 0.0d0
         if (.not. matz) go to 260
c
         do 255 i = 1, n
            t = z(i,k1) + u2 * z(i,k)
            z(i,k1) = z(i,k1) + t * v1
            z(i,k) = z(i,k) + t * v2
  255    continue
c
  260 continue
c     .......... end qz step ..........
      go to 70
c     .......... set error -- all eigenvalues have not
c                converged after 30*n iterations ..........
 1000 ierr = en
c     .......... save epsb for use by qzval and qzvec ..........
 1001 if (n .gt. 1) b(n,1) = epsb
      return
      end

      subroutine qzval(nm,n,a,b,alfr,alfi,beta,matz,z)
c
      integer i,j,n,en,na,nm,nn,isw
      double precision a(nm,n),b(nm,n),alfr(n),alfi(n),beta(n),z(nm,n)
      double precision c,d,e,r,s,t,an,a1,a2,bn,cq,cz,di,dr,ei,ti,tr,u1,
     x       u2,v1,v2,a1i,a11,a12,a2i,a21,a22,b11,b12,b22,sqi,sqr,
     x       ssi,ssr,szi,szr,a11i,a11r,a12i,a12r,a22i,a22r,epsb
      logical matz
c
c     this subroutine is the third step of the qz algorithm
c     for solving generalized matrix eigenvalue problems,
c     siam j. numer. anal. 10, 241-256(1973) by moler and stewart.
c
c     this subroutine accepts a pair of real matrices, one of them
c     in quasi-triangular form and the other in upper triangular form.
c     it reduces the quasi-triangular matrix further, so that any
c     remaining 2-by-2 blocks correspond to pairs of complex
c     eigenvalues, and returns quantities whose ratios give the
c     generalized eigenvalues.  it is usually preceded by  qzhes
c     and  qzit  and may be followed by  qzvec.
c
c     on input
c
c        nm must be set to the row dimension of two-dimensional
c          array parameters as declared in the calling program
c          dimension statement.
c
c        n is the order of the matrices.
c
c        a contains a real upper quasi-triangular matrix.
c
c        b contains a real upper triangular matrix.  in addition,
c          location b(n,1) contains the tolerance quantity (epsb)
c          computed and saved in  qzit.
c
c        matz should be set to .true. if the right hand transformations
c          are to be accumulated for later use in computing
c          eigenvectors, and to .false. otherwise.
c
c        z contains, if matz has been set to .true., the
c          transformation matrix produced in the reductions by qzhes
c          and qzit, if performed, or else the identity matrix.
c          if matz has been set to .false., z is not referenced.
c
c     on output
c
c        a has been reduced further to a quasi-triangular matrix
c          in which all nonzero subdiagonal elements correspond to
c          pairs of complex eigenvalues.
c
c        b is still in upper triangular form, although its elements
c          have been altered.  b(n,1) is unaltered.
c
c        alfr and alfi contain the real and imaginary parts of the
c          diagonal elements of the triangular matrix that would be
c          obtained if a were reduced completely to triangular form
c          by unitary transformations.  non-zero values of alfi occur
c          in pairs, the first member positive and the second negative.
c
c        beta contains the diagonal elements of the corresponding b,
c          normalized to be real and non-negative.  the generalized
c          eigenvalues are then the ratios ((alfr+i*alfi)/beta).
c
c        z contains the product of the right hand transformations
c          (for all three steps) if matz has been set to .true.
c
c     questions and comments should be directed to burton s. garbow,
c     mathematics and computer science div, argonne national laboratory
c
c     this version dated august 1983.
c
c     ------------------------------------------------------------------
c
      epsb = b(n,1)
      isw = 1
c     .......... find eigenvalues of quasi-triangular matrices.
c                for en=n step -1 until 1 do -- ..........
      do 510 nn = 1, n
         en = n + 1 - nn
         na = en - 1
         if (isw .eq. 2) go to 505
         if (en .eq. 1) go to 410
         if (a(en,na) .ne. 0.0d0) go to 420
c     .......... 1-by-1 block, one real root ..........
  410    alfr(en) = a(en,en)
         if (b(en,en) .lt. 0.0d0) alfr(en) = -alfr(en)
         beta(en) = dabs(b(en,en))
         alfi(en) = 0.0d0
         go to 510
c     .......... 2-by-2 block ..........
  420    if (dabs(b(na,na)) .le. epsb) go to 455
         if (dabs(b(en,en)) .gt. epsb) go to 430
         a1 = a(en,en)
         a2 = a(en,na)
         bn = 0.0d0
         go to 435
  430    an = dabs(a(na,na)) + dabs(a(na,en)) + dabs(a(en,na))
     x      + dabs(a(en,en))
         bn = dabs(b(na,na)) + dabs(b(na,en)) + dabs(b(en,en))
         a11 = a(na,na) / an
         a12 = a(na,en) / an
         a21 = a(en,na) / an
         a22 = a(en,en) / an
         b11 = b(na,na) / bn
         b12 = b(na,en) / bn
         b22 = b(en,en) / bn
         e = a11 / b11
         ei = a22 / b22
         s = a21 / (b11 * b22)
         t = (a22 - e * b22) / b22
         if (dabs(e) .le. dabs(ei)) go to 431
         e = ei
         t = (a11 - e * b11) / b11
  431    c = 0.5d0 * (t - s * b12)
         d = c * c + s * (a12 - e * b12)
         if (d .lt. 0.0d0) go to 480
c     .......... two real roots.
c                zero both a(en,na) and b(en,na) ..........
         e = e + (c + dsign(dsqrt(d),c))
         a11 = a11 - e * b11
         a12 = a12 - e * b12
         a22 = a22 - e * b22
         if (dabs(a11) + dabs(a12) .lt.
     x       dabs(a21) + dabs(a22)) go to 432
         a1 = a12
         a2 = a11
         go to 435
  432    a1 = a22
         a2 = a21
c     .......... choose and apply real z ..........
  435    s = dabs(a1) + dabs(a2)
         u1 = a1 / s
         u2 = a2 / s
         r = dsign(dsqrt(u1*u1+u2*u2),u1)
         v1 = -(u1 + r) / r
         v2 = -u2 / r
         u2 = v2 / v1
c
         do 440 i = 1, en
            t = a(i,en) + u2 * a(i,na)
            a(i,en) = a(i,en) + t * v1
            a(i,na) = a(i,na) + t * v2
            t = b(i,en) + u2 * b(i,na)
            b(i,en) = b(i,en) + t * v1
            b(i,na) = b(i,na) + t * v2
  440    continue
c
         if (.not. matz) go to 450
c
         do 445 i = 1, n
            t = z(i,en) + u2 * z(i,na)
            z(i,en) = z(i,en) + t * v1
            z(i,na) = z(i,na) + t * v2
  445    continue
c
  450    if (bn .eq. 0.0d0) go to 475
         if (an .lt. dabs(e) * bn) go to 455
         a1 = b(na,na)
         a2 = b(en,na)
         go to 460
  455    a1 = a(na,na)
         a2 = a(en,na)
c     .......... choose and apply real q ..........
  460    s = dabs(a1) + dabs(a2)
         if (s .eq. 0.0d0) go to 475
         u1 = a1 / s
         u2 = a2 / s
         r = dsign(dsqrt(u1*u1+u2*u2),u1)
         v1 = -(u1 + r) / r
         v2 = -u2 / r
         u2 = v2 / v1
c
         do 470 j = na, n
            t = a(na,j) + u2 * a(en,j)
            a(na,j) = a(na,j) + t * v1
            a(en,j) = a(en,j) + t * v2
            t = b(na,j) + u2 * b(en,j)
            b(na,j) = b(na,j) + t * v1
            b(en,j) = b(en,j) + t * v2
  470    continue
c
  475    a(en,na) = 0.0d0
         b(en,na) = 0.0d0
         alfr(na) = a(na,na)
         alfr(en) = a(en,en)
         if (b(na,na) .lt. 0.0d0) alfr(na) = -alfr(na)
         if (b(en,en) .lt. 0.0d0) alfr(en) = -alfr(en)
         beta(na) = dabs(b(na,na))
         beta(en) = dabs(b(en,en))
         alfi(en) = 0.0d0
         alfi(na) = 0.0d0
         go to 505
c     .......... two complex roots ..........
  480    e = e + c
         ei = dsqrt(-d)
         a11r = a11 - e * b11
         a11i = ei * b11
         a12r = a12 - e * b12
         a12i = ei * b12
         a22r = a22 - e * b22
         a22i = ei * b22
         if (dabs(a11r) + dabs(a11i) + dabs(a12r) + dabs(a12i) .lt.
     x       dabs(a21) + dabs(a22r) + dabs(a22i)) go to 482
         a1 = a12r
         a1i = a12i
         a2 = -a11r
         a2i = -a11i
         go to 485
  482    a1 = a22r
         a1i = a22i
         a2 = -a21
         a2i = 0.0d0
c     .......... choose complex z ..........
  485    cz = dsqrt(a1*a1+a1i*a1i)
         if (cz .eq. 0.0d0) go to 487
         szr = (a1 * a2 + a1i * a2i) / cz
         szi = (a1 * a2i - a1i * a2) / cz
         r = dsqrt(cz*cz+szr*szr+szi*szi)
         cz = cz / r
         szr = szr / r
         szi = szi / r
         go to 490
  487    szr = 1.0d0
         szi = 0.0d0
  490    if (an .lt. (dabs(e) + ei) * bn) go to 492
         a1 = cz * b11 + szr * b12
         a1i = szi * b12
         a2 = szr * b22
         a2i = szi * b22
         go to 495
  492    a1 = cz * a11 + szr * a12
         a1i = szi * a12
         a2 = cz * a21 + szr * a22
         a2i = szi * a22
c     .......... choose complex q ..........
  495    cq = dsqrt(a1*a1+a1i*a1i)
         if (cq .eq. 0.0d0) go to 497
         sqr = (a1 * a2 + a1i * a2i) / cq
         sqi = (a1 * a2i - a1i * a2) / cq
         r = dsqrt(cq*cq+sqr*sqr+sqi*sqi)
         cq = cq / r
         sqr = sqr / r
         sqi = sqi / r
         go to 500
  497    sqr = 1.0d0
         sqi = 0.0d0
c     .......... compute diagonal elements that would result
c                if transformations were applied ..........
  500    ssr = sqr * szr + sqi * szi
         ssi = sqr * szi - sqi * szr
         i = 1
         tr = cq * cz * a11 + cq * szr * a12 + sqr * cz * a21
     x      + ssr * a22
         ti = cq * szi * a12 - sqi * cz * a21 + ssi * a22
         dr = cq * cz * b11 + cq * szr * b12 + ssr * b22
         di = cq * szi * b12 + ssi * b22
         go to 503
  502    i = 2
         tr = ssr * a11 - sqr * cz * a12 - cq * szr * a21
     x      + cq * cz * a22
         ti = -ssi * a11 - sqi * cz * a12 + cq * szi * a21
         dr = ssr * b11 - sqr * cz * b12 + cq * cz * b22
         di = -ssi * b11 - sqi * cz * b12
  503    t = ti * dr - tr * di
         j = na
         if (t .lt. 0.0d0) j = en
         r = dsqrt(dr*dr+di*di)
         beta(j) = bn * r
         alfr(j) = an * (tr * dr + ti * di) / r
         alfi(j) = an * t / r
         if (i .eq. 1) go to 502
  505    isw = 3 - isw
  510 continue
      b(n,1) = epsb
c
      return
      end

      subroutine qzvec(nm,n,a,b,alfr,alfi,beta,z)
c
      integer i,j,k,m,n,en,ii,jj,na,nm,nn,isw,enm2
      double precision a(nm,n),b(nm,n),alfr(n),alfi(n),beta(n),z(nm,n)
      double precision d,q,r,s,t,w,x,y,di,dr,ra,rr,sa,ti,tr,t1,t2,w1,x1,
     x       zz,z1,alfm,almi,almr,betm,epsb
c
c     this subroutine is the optional fourth step of the qz algorithm
c     for solving generalized matrix eigenvalue problems,
c     siam j. numer. anal. 10, 241-256(1973) by moler and stewart.
c
c     this subroutine accepts a pair of real matrices, one of them in
c     quasi-triangular form (in which each 2-by-2 block corresponds to
c     a pair of complex eigenvalues) and the other in upper triangular
c     form.  it computes the eigenvectors of the triangular problem and
c     transforms the results back to the original coordinate system.
c     it is usually preceded by  qzhes,  qzit, and  qzval.
c
c     on input
c
c        nm must be set to the row dimension of two-dimensional
c          array parameters as declared in the calling program
c          dimension statement.
c
c        n is the order of the matrices.
c
c        a contains a real upper quasi-triangular matrix.
c
c        b contains a real upper triangular matrix.  in addition,
c          location b(n,1) contains the tolerance quantity (epsb)
c          computed and saved in  qzit.
c
c        alfr, alfi, and beta  are vectors with components whose
c          ratios ((alfr+i*alfi)/beta) are the generalized
c          eigenvalues.  they are usually obtained from  qzval.
c
c        z contains the transformation matrix produced in the
c          reductions by  qzhes,  qzit, and  qzval, if performed.
c          if the eigenvectors of the triangular problem are
c          desired, z must contain the identity matrix.
c
c     on output
c
c        a is unaltered.  its subdiagonal elements provide information
c           about the storage of the complex eigenvectors.
c
c        b has been destroyed.
c
c        alfr, alfi, and beta are unaltered.
c
c        z contains the real and imaginary parts of the eigenvectors.
c          if alfi(i) .eq. 0.0, the i-th eigenvalue is real and
c            the i-th column of z contains its eigenvector.
c          if alfi(i) .ne. 0.0, the i-th eigenvalue is complex.
c            if alfi(i) .gt. 0.0, the eigenvalue is the first of
c              a complex pair and the i-th and (i+1)-th columns
c              of z contain its eigenvector.
c            if alfi(i) .lt. 0.0, the eigenvalue is the second of
c              a complex pair and the (i-1)-th and i-th columns
c              of z contain the conjugate of its eigenvector.
c          each eigenvector is normalized so that the modulus
c          of its largest component is 1.0 .
c
c     questions and comments should be directed to burton s. garbow,
c     mathematics and computer science div, argonne national laboratory
c
c     this version dated august 1983.
c
c     ------------------------------------------------------------------
c
      epsb = b(n,1)
      isw = 1
c     .......... for en=n step -1 until 1 do -- ..........
      do 800 nn = 1, n
         en = n + 1 - nn
         na = en - 1
         if (isw .eq. 2) go to 795
         if (alfi(en) .ne. 0.0d0) go to 710
c     .......... real vector ..........
         m = en
         b(en,en) = 1.0d0
         if (na .eq. 0) go to 800
         alfm = alfr(m)
         betm = beta(m)
c     .......... for i=en-1 step -1 until 1 do -- ..........
         do 700 ii = 1, na
            i = en - ii
            w = betm * a(i,i) - alfm * b(i,i)
            r = 0.0d0
c
            do 610 j = m, en
  610       r = r + (betm * a(i,j) - alfm * b(i,j)) * b(j,en)
c
            if (i .eq. 1 .or. isw .eq. 2) go to 630
            if (betm * a(i,i-1) .eq. 0.0d0) go to 630
            zz = w
            s = r
            go to 690
  630       m = i
            if (isw .eq. 2) go to 640
c     .......... real 1-by-1 block ..........
            t = w
            if (w .eq. 0.0d0) t = epsb
            b(i,en) = -r / t
            go to 700
c     .......... real 2-by-2 block ..........
  640       x = betm * a(i,i+1) - alfm * b(i,i+1)
            y = betm * a(i+1,i)
            q = w * zz - x * y
            t = (x * s - zz * r) / q
            b(i,en) = t
            if (dabs(x) .le. dabs(zz)) go to 650
            b(i+1,en) = (-r - w * t) / x
            go to 690
  650       b(i+1,en) = (-s - y * t) / zz
  690       isw = 3 - isw
  700    continue
c     .......... end real vector ..........
         go to 800
c     .......... complex vector ..........
  710    m = na
         almr = alfr(m)
         almi = alfi(m)
         betm = beta(m)
c     .......... last vector component chosen imaginary so that
c                eigenvector matrix is triangular ..........
         y = betm * a(en,na)
         b(na,na) = -almi * b(en,en) / y
         b(na,en) = (almr * b(en,en) - betm * a(en,en)) / y
         b(en,na) = 0.0d0
         b(en,en) = 1.0d0
         enm2 = na - 1
         if (enm2 .eq. 0) go to 795
c     .......... for i=en-2 step -1 until 1 do -- ..........
         do 790 ii = 1, enm2
            i = na - ii
            w = betm * a(i,i) - almr * b(i,i)
            w1 = -almi * b(i,i)
            ra = 0.0d0
            sa = 0.0d0
c
            do 760 j = m, en
               x = betm * a(i,j) - almr * b(i,j)
               x1 = -almi * b(i,j)
               ra = ra + x * b(j,na) - x1 * b(j,en)
               sa = sa + x * b(j,en) + x1 * b(j,na)
  760       continue
c
            if (i .eq. 1 .or. isw .eq. 2) go to 770
            if (betm * a(i,i-1) .eq. 0.0d0) go to 770
            zz = w
            z1 = w1
            r = ra
            s = sa
            isw = 2
            go to 790
  770       m = i
            if (isw .eq. 2) go to 780
c     .......... complex 1-by-1 block ..........
            tr = -ra
            ti = -sa
  773       dr = w
            di = w1
c     .......... complex divide (t1,t2) = (tr,ti) / (dr,di) ..........
  775       if (dabs(di) .gt. dabs(dr)) go to 777
            rr = di / dr
            d = dr + di * rr
            t1 = (tr + ti * rr) / d
            t2 = (ti - tr * rr) / d
            go to (787,782), isw
  777       rr = dr / di
            d = dr * rr + di
            t1 = (tr * rr + ti) / d
            t2 = (ti * rr - tr) / d
            go to (787,782), isw
c     .......... complex 2-by-2 block ..........
  780       x = betm * a(i,i+1) - almr * b(i,i+1)
            x1 = -almi * b(i,i+1)
            y = betm * a(i+1,i)
            tr = y * ra - w * r + w1 * s
            ti = y * sa - w * s - w1 * r
            dr = w * zz - w1 * z1 - x * y
            di = w * z1 + w1 * zz - x1 * y
            if (dr .eq. 0.0d0 .and. di .eq. 0.0d0) dr = epsb
            go to 775
  782       b(i+1,na) = t1
            b(i+1,en) = t2
            isw = 1
            if (dabs(y) .gt. dabs(w) + dabs(w1)) go to 785
            tr = -ra - x * b(i+1,na) + x1 * b(i+1,en)
            ti = -sa - x * b(i+1,en) - x1 * b(i+1,na)
            go to 773
  785       t1 = (-r - zz * b(i+1,na) + z1 * b(i+1,en)) / y
            t2 = (-s - zz * b(i+1,en) - z1 * b(i+1,na)) / y
  787       b(i,na) = t1
            b(i,en) = t2
  790    continue
c     .......... end complex vector ..........
  795    isw = 3 - isw
  800 continue
c     .......... end back substitution.
c                transform to original coordinate system.
c                for j=n step -1 until 1 do -- ..........
      do 880 jj = 1, n
         j = n + 1 - jj
c
         do 880 i = 1, n
            zz = 0.0d0
c
            do 860 k = 1, j
  860       zz = zz + z(i,k) * b(k,j)
c
            z(i,j) = zz
  880 continue
c     .......... normalize so that modulus of largest
c                component of each vector is 1.
c                (isw is 1 initially from before) ..........
      do 950 j = 1, n
         d = 0.0d0
         if (isw .eq. 2) go to 920
         if (alfi(j) .ne. 0.0d0) go to 945
c
         do 890 i = 1, n
            if (dabs(z(i,j)) .gt. d) d = dabs(z(i,j))
  890    continue
c
         do 900 i = 1, n
  900    z(i,j) = z(i,j) / d
c
         go to 950
c
  920    do 930 i = 1, n
            r = dabs(z(i,j-1)) + dabs(z(i,j))
            if (r .ne. 0.0d0) r = r * dsqrt((z(i,j-1)/r)**2
     x                                     +(z(i,j)/r)**2)
            if (r .gt. d) d = r
  930    continue
c
         do 940 i = 1, n
            z(i,j-1) = z(i,j-1) / d
            z(i,j) = z(i,j) / d
  940    continue
c
  945    isw = 3 - isw
  950 continue
c
      return
      end

c----------------------------------------------------------------------c
c                          S P A R S K I T                             c
c----------------------------------------------------------------------c

c
c     real*8 function pddot(n,x,ix,y,iy)
c     integer n, ix, iy
c     real*8 x(1+(n-1)*ix), y(1+(n-1)*iy)
c
c     This interface of pddot is exactly the same as that of
c     DDOT (or SDOT if real == real*8) from BLAS-1. It should have
c     same functionality as DDOT on a single processor machine. On a
c     parallel/distributed environment, each processor can perform
c     DDOT on the data it has, then perform a summation on all the
c     partial results.
c

      real(kind=8) function pddot(n,x,ix,y,iy)
      include 'mpif.h'
      integer n, ix, iy
      real(kind=8) lsd, x(*), y(*), ddot
      external ddot

      lsd = ddot(n,x,ix,y,iy)
      call MPI_allreduce(lsd,pddot,1,
     *     MPI_DOUBLE_PRECISION,
     *     MPI_SUM,MPI_COMM_WORLD, ierr)
      return
      end
c-----end-of-pddot


